"""Aliases for functions. This module is not guaranteed to be stable."""

from aind_mri_utils.file_io import simpleitk as sitk_io
from aind_mri_utils.rotations import (
    create_homogeneous_from_euler_and_translation,
    prepare_data_for_homogeneous_transform,
)

append_ones_columns = prepare_data_for_homogeneous_transform
create_rigid_transform = create_homogeneous_from_euler_and_translation


def save_sitk_transform(filename, T, transpose_matrix=False):
    """
    This is an alias for `sitk_io.save_sitk_transform` that has the same
    interface as the original function that Yoni wrote.
    """

    sitk_io.save_sitk_transform(
        filename,
        rotation_matrix=T,
        transpose_matrix=transpose_matrix,
        legacy=True,
    )



"""Module to fit implant rotations to MRI data."""

import warnings
from concurrent.futures import ProcessPoolExecutor, as_completed

import numpy as np
from scipy.optimize import fmin
from scipy.spatial.transform import Rotation

from aind_mri_utils.file_io.slicer_files import get_segmented_labels
from aind_mri_utils.meshes import (
    distance_to_all_triangles_in_mesh,
    distance_to_closest_point_for_each_triangle_in_mesh,
)
from aind_mri_utils.rotations import (
    apply_rotate_translate,
    combine_angles,
    rotation_matrix_from_vectors,
)
from aind_mri_utils.sitk_volume import find_points_equal_to


def _params_to_rotation_translation(params):
    """
    Convert parameters to rotation matrix and translation vector.

    Parameters
    ----------
    params : array-like
        Parameters for optimization including Euler angles and translation
        vector.

    Returns
    -------
    ndarray
        Rotation matrix.
    ndarray
        Translation vector.
    """
    rotation_matrix = combine_angles(*params[:3])
    translation = params[3:]
    return rotation_matrix, translation


def _implant_cost_fun(T, hole_mesh_dict, hole_seg_dict, run_parallel=True):
    """
    Computes the total distance cost for implant alignment based on the
    provided transformation parameters.

    Parameters
    ----------
    T : array-like
        Transformation parameters including Euler angles and translation
        vector.
    hole_mesh_dict : dict
        Dictionary where keys are hole IDs and values are mesh objects
        representing the holes.
    hole_seg_dict : dict
        Dictionary where keys are hole IDs and values are segmented points
        corresponding to the holes.

    Returns
    -------
    float
        The total distance cost calculated by summing the distances between
        transformed points and mesh triangles.
    """
    rotation_matrix, translation = _params_to_rotation_translation(T)
    tasks = []
    for hole_id in hole_mesh_dict.keys():
        # TODO: Fix this so it actually works for the brain outline
        if hole_id not in hole_seg_dict:
            continue
        mesh = hole_mesh_dict[hole_id].copy()
        pts = hole_seg_dict[hole_id]
        mesh.vertices = apply_rotate_translate(
            mesh.vertices, rotation_matrix, translation
        )
        args = (mesh, pts)
        if hole_id == -1:
            func = distance_to_closest_point_for_each_triangle_in_mesh
        else:
            func = distance_to_all_triangles_in_mesh
        tasks.append((func, args))
    total_distance = 0.0
    if run_parallel:
        with ProcessPoolExecutor() as executor:
            futures = [executor.submit(func, *args) for func, args in tasks]
            for future in as_completed(futures):
                distances, _ = future.result()
                total_distance += np.sum(distances)
    else:
        for func, args in tasks:
            distances, _ = func(*args)
            total_distance += np.sum(distances)
    return total_distance


def fit_implant_to_mri(
    hole_seg_dict,
    hole_mesh_dict,
    initialization_hole=4,
    other_init_holes=[3, 9],
):
    """
    Fits an implant model to MRI data by optimizing the alignment of hole
    segments.

    Parameters
    ----------
    hole_seg_dict : dict
        Dictionary containing segmented hole data from MRI. Keys are hole
        identifiers, and values are numpy arrays of coordinates.
    hole_mesh_dict : dict
        Dictionary containing mesh data for the implant model. Keys are hole
        identifiers, and values are mesh objects with vertex coordinates. Lower
        face has key -1.
    initialization_hole : int, optional
        The hole to use for initialization, by default 4.
    other_init_holes : int, optional
        The hole to use for initialization of the rotation, by default [3, 9].


    Returns
    -------
    rotation_matrix : ndarray
        The rotation matrix that aligns the implant model to the MRI data.
    translation : ndarray
        The translation vector that aligns the implant model to the MRI data.
    """
    T = np.zeros(6)
    initialize_translation = initialization_hole in hole_seg_dict
    if initialize_translation:
        annotation_mean = np.mean(hole_seg_dict[initialization_hole], axis=0)
        model_mean = np.mean(
            hole_mesh_dict[initialization_hole].vertices, axis=0
        )
        init_translation = annotation_mean - model_mean
        T[3:] = init_translation
    else:
        warnings.warn(
            f"Could not find hole {initialization_hole} in MRI data "
            "for initialization"
        )

    # Initial guess of rotation
    initialize_rotation = initialize_translation and len(other_init_holes) >= 2
    for other_hole in other_init_holes:
        initialize_rotation = (
            initialize_rotation
            and other_hole in hole_seg_dict
            and other_hole != initialization_hole
        )
        if not initialize_rotation:
            break

    if initialize_rotation:
        # Initialize rotation based on pairs of holes
        #
        # Find the rotation that aligns the vector between the first two holes
        # in the annotation to the vector between the first two holes in the
        # model. Refine with subsequent pairs of holes.
        R_init = np.eye(3)
        for other_hole in other_init_holes:
            other_anno_mean = np.mean(hole_seg_dict[other_hole], axis=0)
            other_model_mean = np.mean(
                hole_mesh_dict[other_hole].vertices, axis=0
            )
            model_diff_rotated = R_init @ (other_model_mean - model_mean)
            R_update = rotation_matrix_from_vectors(
                model_diff_rotated, other_anno_mean - annotation_mean
            )
            R_init = R_update @ R_init
        T[:3] = Rotation.from_matrix(R_init).as_euler("xyz")
    else:
        warnings.warn("Could not find holes for rotation initialization")

    output = fmin(
        _implant_cost_fun,
        T,
        args=(hole_mesh_dict, hole_seg_dict),
        xtol=1e-6,
        maxiter=2000,
    )
    rotation_matrix, translation = _params_to_rotation_translation(output)
    return rotation_matrix, translation


def make_hole_seg_dict(implant_annotations):
    """
    Creates a dictionary mapping hole names to their segmented positions.

    Parameters
    ----------
    implant_annotations : numpy.ndarray
        An array containing the implant annotations.

    Returns
    -------
    dict
        A dictionary where the keys are hole names (as integers) and the values
        are lists of positions where the segmented values are found.
    """
    # TODO: Fix this so it actually works for the brain outline
    implant_annotations_names = get_segmented_labels(implant_annotations)
    hole_seg_dict = {}
    for hole_name, seg_val in implant_annotations_names.items():
        positions = find_points_equal_to(implant_annotations, seg_val)
        hole_seg_dict[int(hole_name)] = positions
    return hole_seg_dict



# -*- coding: utf-8 -*-
"""
Functions for Loading and manipulating meshes during insertion planning.
"""

import numpy as np
import trimesh

from aind_mri_utils.rotations import make_homogeneous_transform


def as_mesh(scene_or_mesh):
    """
    Convert a possible scene to a mesh.

    If conversion occurs, the returned mesh has only vertex and face data.

    see https://github.com/mikedh/trimesh/issues/507
    """
    if isinstance(scene_or_mesh, trimesh.Scene):
        if len(scene_or_mesh.geometry) == 0:
            mesh = None  # empty scene
        else:
            # we lose texture information here
            mesh = trimesh.util.concatenate(
                tuple(
                    trimesh.Trimesh(vertices=g.vertices, faces=g.faces)
                    for g in scene_or_mesh.geometry.values()
                )
            )
    else:
        assert isinstance(scene_or_mesh, trimesh.Trimesh)
        mesh = scene_or_mesh
    return mesh


def load_newscale_trimesh(
    filename,
    move_down=0,
):
    """
    Load a newscale model mesh

    """
    mesh = trimesh.load_mesh(filename)
    mesh = as_mesh(mesh)
    mesh.vertices = mesh.vertices[:, [0, 2, 1]]  # Should be made into params.
    mesh.vertices[:, 2] = mesh.vertices[:, 2] - move_down
    # Repair broken stuff from bad blender-ing
    trimesh.repair.broken_faces(mesh)
    trimesh.repair.fix_normals(mesh)
    trimesh.repair.fix_inversion(mesh)
    trimesh.repair.fix_winding(mesh)
    return mesh


def apply_transform_to_trimesh(mesh, R, translation=None):
    """
    Apply a transform to a trimesh Mesh object
    """
    if translation is None:
        translation = np.zeros(3)
    T = make_homogeneous_transform(R, translation)
    mesh.vertices = trimesh.transform_points(mesh.vertices, T)
    return mesh


def create_uv_spheres(positions, radius=0.25, color=[255, 0, 255, 255]):
    """
    Create UV spheres at specified positions with a given radius and color.

    Parameters
    ----------
    positions : ndarray
        Array of positions where spheres should be created.
    radius : float, optional
        Radius of the spheres, by default 0.25.
    color : list, optional
        RGBA color of the spheres, by default [255, 0, 255, 255].

    Returns
    -------
    list
        List of trimesh objects representing the spheres.
    """
    meshes = [
        trimesh.creation.uv_sphere(radius=radius)
        for _ in range(len(positions))
    ]
    for i, mesh in enumerate(meshes):
        mesh.apply_translation(positions[i, :])
        mesh.visual.vertex_colors = color
    return meshes


def distances_to_triangle(points, triangle):
    """
    Calculate the distance from a set of points to a triangle.

    Parameters
    ----------
    points : array-like, shape (n, 3)
        The points from which the distance to the triangle is calculated. Each
        row represents a point with x, y, z coordinates.
    triangle : array-like, shape (3, 3)
        The vertices of the triangle. Each row represents a vertex with x, y, z
        coordinates.

    Returns
    -------
    distances : numpy.ndarray, shape (n,)
        The distances from each point to the triangle.
    nearest_points : numpy.ndarray, shape (n, 3)
        The nearest points on the triangle for each input point.
    """
    tri_mesh = trimesh.Trimesh(vertices=triangle, faces=[[0, 1, 2]])
    nearest_points, distances, _ = trimesh.proximity.closest_point(
        tri_mesh, points
    )
    return distances, nearest_points


def distance_to_all_triangles_in_mesh(mesh, points, normalize=True):
    """Calculate the distances from points to all triangles in a mesh.

    This function computes the minimum distance between a set of points and
    each triangle in a mesh, along with the nearest points on the triangles.

    Parameters
    ----------
    mesh : trimesh.Trimesh
        A triangular mesh object containing the triangles to measure distance
        from
    points : numpy.ndarray
        Array of points to calculate distances from, shape (N, 3) where N is
        number of points
    normalize : boolean, optional
        Whether to normalize distances by number of points, by default True

    Returns
    -------
    distances : numpy.ndarray
        Array of distances from points to each triangle in the mesh
        If normalize=True, distances are divided by number of points
    nearest_points : list
        List of arrays containing the nearest points on each triangle for each
        input point

    See Also
    --------
    distance_to_triangle : Function that calculates distance from points to a
    single triangle
    """
    distances = []
    nearest_points = []
    for triangle in mesh.triangles:
        this_distance, this_nearest_points = distances_to_triangle(
            points, triangle
        )
        distances.append(this_distance)
        nearest_points.append(this_nearest_points)
    distances = np.array(distances)
    if normalize:
        distances = distances / len(points)
    return distances, nearest_points


def distance_to_closest_point_for_each_triangle_in_mesh(
    mesh, points, normalize=True
):
    """
    Calculate the distance to the closest point for each triangle in a mesh.

    Parameters
    ----------
    mesh : object
        A mesh object that contains triangles.
    points : array-like
        An array of points to calculate the distance from.
    normalize : boolean, optional
        If set to True, the distances will be normalized by the number of
        triangles (default is True).

    Returns
    -------
    distances : numpy.ndarray
        An array of the minimum distances from each triangle to the closest
        point.
    nearest_points : list of array-like
        A list of the nearest points corresponding to each triangle.

    Notes
    -----
    This function assumes that the `mesh` object has an attribute `triangles`
    which is an array of triangles.  The `distance_to_triangle` function is
    used to calculate the distance from a point to a triangle.
    """
    triangles = mesh.triangles
    distances = []
    nearest_points = []
    for triangle in triangles:
        distances_to_tri, nearest_points_tri = distances_to_triangle(
            points, triangle
        )
        min_ndx = np.argmin(distances_to_tri)
        distances.append(distances_to_tri[min_ndx])
        nearest_points.append(nearest_points_tri[min_ndx, :])
    distances = np.array(distances)
    if normalize:
        distances = distances / len(triangles)
    return distances, nearest_points
